/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>

extern int logfd;
extern void log(const char *fmt, ...);
extern void fatal(const char *fmt, ...);

void
daemonize(const char *cmd)
{
	int i;
	pid_t pid;
	struct rlimit rl;
	char filename[256];

	/*
	 * Clear file creation mask.
	 */
	umask(0);

	/*
	 * Create the log file.
	 */
	sprintf(filename, "/tmp/%s.log", cmd);
	if (initlog(filename) < 0)
		fatal("%s: cannot create log file", cmd);

	/*
	 * Get the maximum number of files we can open.
	 */
	if (getrlimit(RLIMIT_NOFILE, &rl) < 0)
		fatal("%s: cannot get file limit", cmd);

	/*
	 * Become a session leader to lose our
	 * controlling terminal.
	 */
	if ((pid = fork()) < 0)
		fatal("%s: cannot fork", cmd);
	else if (pid != 0)	/* parent */
		exit(0);
	setsid();

	/*
	 * Ensure that future opens won't allocate
	 * controlling terminals.
	 */
	sigset(SIGHUP, SIG_IGN);
	if ((pid = fork()) < 0)
		fatal("%s: cannot fork", cmd);
	else if (pid != 0)	/* parent */
		exit(0);

	/*
	 * Close all open file descriptors, except for
	 * the log file.  Choose a reasonable limit if
	 * the limit is currently infinite.
	 */
	if (rl.rlim_max == RLIM_INFINITY)
		rl.rlim_max = 1024;
	for (i = 0; i < rl.rlim_max; i++)
		if (i != logfd)
			close(i);
	errno = 0;	/* set from close of invalid fd */

	/*
	 * Change the current directory to the root of
	 * the file system tree so other file systems can
	 * be unmounted while we're running.
	 */
	if (chdir("/") < 0) {
		log("cannot chdir to /: %s", strerror(errno));
		exit(1);
	}

	/*
	 * Attach file descriptors 0, 1, and 2 to /dev/null.
	 * Applications expect them to be at least open.
	 */
	if ((i = open("/dev/null", O_RDWR)) != 0) {
		if (i < 0)
			log("error opening /dev/null: %s",
			  strerror(errno));
		else
			log("open /dev/null: expected fd 0, got %d",
			  i);
		exit(1);
	}
	if ((i = dup(0)) != 1) {
		if (i < 0)
			log("dup failed: %s", strerror(errno));
		else
			log("dup(0) expected fd 1, got %d", i);
		exit(1);
	}
	if ((i = dup(0)) != 2) {
		if (i < 0)
			log("dup failed: %s", strerror(errno));
		else
			log("dup(0) expected fd 2, got %d", i);
		exit(1);
	}
}
